/*-------------------------------------------------
 *	Real Time Linux version 3
 *	Kernel version 2.2.x - 2.4.20
 *	RS485 (Moxa132, Moxa102, COM-port, LPT-port :
 * 
 *	Andre Zaitsev, Olga Voziakova
 *	ovoz@sai.msu.ru
 *	$History:
 *
 *---------------------------------------- */
 
#include <linux/config.h> 
 
 /*    */
#if CONFIG_MODVERSIONS==1
#include <linux/modversions.h>
#endif

#include <linux/module.h>
#include <linux/version.h>

#include <linux/errno.h>
#include <linux/major.h>
#include <linux/slab.h>
#include <asm/uaccess.h>	//  copy_from_user
#include <linux/ioport.h>
#include <linux/delay.h>	// udelay

#include <mcua.h>
#include <rserrno.h>

#ifdef __CP132__
	#include <moxa132.h>
#endif
#ifdef __CP102__
	#include <moxa102.h>
#endif
#ifdef __COM__
	#include <com.h>
#endif
#ifdef __LPT__
	#include <lpt.h>
#endif

#include <rs485.h>

extern int		rs485_request_irq(unsigned int irq);
extern void		rs485_clean_irq(unsigned int irq);
extern int		start_transmitter(RS485INFO *info);

extern int		rs485_fifo_create(RS485INFO *info);
extern void		rs485_fifo_init(struct file* file);
extern void		rs485_fifo_destroy(void);
extern void		rs485_fifo_flush(RS485INFO *info);
extern int		rs485_data_fifo_test( int count, RS485INFO* info );
extern int		rs485_data_fifo_put( void* buf,
								int count, RS485INFO* info);
extern ssize_t	rs485_data_fifo_read( RS485INFO* info,
								char* buf, size_t count);

#ifdef __RS485_PROC__
extern int rs485_create_proc(void);
extern void rs485_remove_proc(void);
#endif

// ,     insmod
static int io1 = 0;			//  1  -
static int io2 = 0;			//  2  -
static int irq1 = 0;			//   port1
static int irq2 = 0;			//   port2
//static int portnumb = 0;	//  : 1-COM1 ( PORT1  MOXA102),
							// 2-COM2 ( PORT2  MOXA102)...
static int baudrate1 = 0;		//   port1
static int baudrate2 = 0;		//   port2

#ifndef __PCI__	//   ISA ( Moxa C102  COM-)
  extern int rs485_getIO(int port);
  extern int rs485_getIRQ(int port);
#else
	extern  int _find_PCImoxa(void);
#endif

//extern int rs485_init(void);
extern int rs485_set_baudrate(RS485INFO *pinfo);
extern void rs485_reset(RS485INFO *pinfo);
//extern inline int rs485_start(RS485INFO *pinfo);
extern int rs485_probe(RS485INFO *pinfo);
extern inline void rs485_allow_interrupt_card(RS485INFO *pinfo);
extern inline void rs485_stop_interrupt_card(RS485INFO *pinfo);
extern inline int rs485_test_device(RS485INFO *pinfo);
#ifndef __LPT__
	extern inline void	rs485_clean_reg(RS485INFO *info);
#endif

extern inline int rs485_timeout(long interval);
extern inline void rs485_set_starttime(void);

struct file_operations	rs485_fops;
static RS485INFO		*info_port1, *info_port2;

#ifndef __LPT__
char crc(unsigned char *buf, unsigned short size) {

/*-------------------------------------------------
 *	    
 *   
 *  *buf -    
 *  size -  
 --------------------------------------------------*/	

	unsigned short acc;
    unsigned short divisor = 0x700;
    short i,j;

    *(buf+size) = 0x0;
    acc = (*(buf))<<8;
    for( j=0; j<size; j++ ){ 
        acc = acc|(*(buf+j+1));
        for( i=0; i< 8; i++ ){
            if( acc&0x8000 ) {
                acc = acc<<1;
                acc = acc^divisor;
            } else acc = acc<<1;
        }
    }
    acc = acc>>8;
    *(buf+size) = (char)acc;
    return( (char)acc );
}
#endif

int rs485_init(void) {		//  
	int major = RS485_MAJOR;
	
#ifdef __PCI__
	int res = _find_PCImoxa();
	if(res < SUCCESS) return res;
#endif

	if(check_region(io1, ADDRESS_REGION_SIZE)) {
		RS_PRINT( KERN_CRIT "%s: I/O port 0x%08lX in use.\n",
					ADAPTER_ID, (unsigned long) io1);
		io1 = 0;
//		return(-EACCES);
	} else {
		request_region(io1, ADDRESS_REGION_SIZE, ADAPTER_ID);
	}
	
	if(check_region(io2, ADDRESS_REGION_SIZE)) {
		if( io1 != 0 ) release_region(io1, ADDRESS_REGION_SIZE);
		RS_PRINT( KERN_CRIT "%s: I/O port 0x%08lX in use.\n",
					ADAPTER_ID, (unsigned long) io2);
		io2 = 0;
//		return(-EACCES);
	} else {
		request_region( io2, ADDRESS_REGION_SIZE, ADAPTER_ID );
	}	

	if( (io1 == 0) && (io2 == 0) ) return(-EACCES);
	if(register_chrdev (major, ADAPTER_ID, &rs485_fops)) {
		RS_PRINT( KERN_CRIT "%s: unable to get major %d\n", ADAPTER_ID, major);
		if( io2 != 0 ) release_region(io2, ADDRESS_REGION_SIZE);
		if( io1 != 0 ) release_region(io1, ADDRESS_REGION_SIZE);
		return (-EIO);
	}

	return(SUCCESS);
}

int rs485_put_data(RS485INFO *rs485) {

//  active   FIFO
// Proiskhodit v obrabotchike preryvaniya => nikto vklinit'sia ne smozhet

//VK	unsigned int dev, idev;
	unsigned int idev;
	unsigned char c_dev;
	int num;
	int err;

//VK	dev = ((unsigned int) rs485->active_inbuf.adr & MASK_ADR_PACKET);	//  
//VK	idev = dev-1;
	idev = ((unsigned int) rs485->active_inbuf.adr & MASK_ADR_PACKET);	//  
	num = rs485->active_inbuf.adr & MASK_CNT_PACKET;	//   
	num /= 0x20; 	
//    
	if( rs485->device[idev].round_count_prev == num ) {	//    
		rs485->active_inbuf.signature = 0;
		return(-1);
	}

	if( num < rs485->device[idev].round_count_prev ) {
		(rs485->device[idev].number) += 4;
	}

	rs485->device[idev].number &= ~3;
	rs485->device[idev].number |= num;
	rs485->device[idev].round_count_prev = num;

#ifndef __LPT__
	rs485->active_inbuf.size ++;	//    
#else
//  CRC
	rs485->active_inbuf.size += 2;	//    
#endif
	/* ----------------------------------------
	 *     
	 * +0	 
	 * +1	 
	 * +2	 
	 * +3	  (2)
	 * +5	 
	 ------------------------------------------ */

	if( rs485_data_fifo_test(rs485->active_inbuf.size+1, rs485) < 0 ) {
		PDEBUG( "rs485_put_data_fifo: FIFO overflow\n" );
		rs485_fifo_flush(rs485);
		return -ERSFIFOVERFLOW;
	}

	err = rs485_data_fifo_put( &rs485->active_inbuf.size,
					sizeof(unsigned char), rs485 );
	if( err < 0 ) {
		PDEBUG( "rs485_put_data_fifo: err-code=%d\n",err );
		return err;
	}

	err = rs485_data_fifo_put( &rs485->active_inbuf.signature,
					sizeof(unsigned char), rs485 );

	if( err < 0 ) {
		PDEBUG( "rs485_put_data_fifo: err-code=%d\n",err );
		return err;
	}

	rs485->active_inbuf.signature = 0;
//VK	c_dev = (unsigned char)dev;
	c_dev = (unsigned char)idev;
	err = rs485_data_fifo_put( &c_dev,
					sizeof(unsigned char), rs485 );
	if( err < 0 ) {
		PDEBUG( "rs485_put_data_fifo: err-code=%d\n",err );
		return err;
	}
	err = rs485_data_fifo_put( &rs485->device[idev].number,
					sizeof(unsigned short), rs485 );
	if( err < 0 ) {
		PDEBUG( "rs485_put_data_fifo: err-code=%d\n",err );
		return err;
	}
	err = rs485_data_fifo_put( &(rs485->active_inbuf.cache[2]),
							rs485->active_inbuf.size-4, rs485 );

	if( err <= 0 ) {
		PDEBUG( "rs485_put_data_fifo: err-code=%d\n",err );
		return err;
	}

	return(SUCCESS);
}

void rs485_blnumber_reset(RS485INFO *info) {
//	  
	int i;
//VK	for( i=0; i<MAX_TASK_DEVICE+1; i++ ) {	
	for( i=0; i<MAX_TASK_DEVICE; i++ ) {	
		info->device[i].number = 0;
		info->device[i].round_count_prev = -1;
	}
	info->passive_inbuf.adr = 0;
	info->active_inbuf.adr = 0;
}

/* ------------------------------------------------------------
 *    
 * -----------------------------------------------------------*/

static int rs485_open(struct inode *inode, struct file *file) {

//    .     

	RS485INFO *info;
	int ret=0;
	
	switch(MINOR(inode->i_rdev)) {		//   
		case 0:
			if( io1 == 0 ) return(-ERSBUSY);
			info = info_port1;			//   1
			break;

		case 1:
			if( io2 == 0) return(-ERSBUSY);
			info = info_port2;			//   2
			break;

		default: return(-ERSNOPORT);
	}

	file->private_data = info;

	PDEBUG("%s: open device %d, %d\n",
			ADAPTER_ID, MAJOR(inode->i_rdev), MINOR(inode->i_rdev));

	if( info->use_mcua == 0 ) {			//   

		rs485_fifo_init(file);			//  FIFO   
		ret = rs485_test_device(info);
		if( ret < 0 ) return(ret);
		rs485_reset(info);

	} else if( info->use_mcua >= MAX_USE_STREAM ) {
// 		    
		return(-ERSBUSY);
	}

	if( test_flags(FMCUA_OPEN, &info->lFlags) == 0 ) {
		up_flags(FMCUA_OPEN, &info->lFlags);
	}

	info->use_mcua ++;
	MOD_INC_USE_COUNT;
	return(SUCCESS);	
}

static int rs485_close(struct inode *inode, struct file *file) {
/*     
 *      
 */

	RS485INFO *info = file->private_data;
	MOD_DEC_USE_COUNT;
	info->use_mcua --;

	if( info->use_mcua == 0 ) {

		//      
		rs485_stop_interrupt_card(info);
#ifndef __LPT__
		rs485_clean_reg(info);
#endif
    }
	down_flags(FMCUA_OPEN, &info->lFlags);

	return(SUCCESS);
}

static ssize_t rs485_write( struct file *fp, const char *buf, size_t size, loff_t *offset ) {

/*      ( NE   write )
 *    < 0  	 
 *				> CACHE_MAXSIZE	  
 *				< CACHE_MAXSIZE	   ,
 * 					zapisyvaemykh v pol'zovatel'skij buf
 *				CRC ne peredayotsia pol'zovateliu.
 */	

	RS485INFO	*info = fp->private_data;
#ifndef __LPT__
	unsigned char crc_sum;
#endif	
	int ret = 0;

	if( size > CACHE_MAXSIZE ) {
		PDEBUG( "write: size=%d is too match!\n", size);
		return(-ERSFORMAT);
	}

	rs485_set_starttime();
	while(1) {
		if( !test_flags(FRECIVER, &info->lFlags)
				&& !test_flags(FTRANSMITTER, &info->lFlags) ) break;
		if( rs485_timeout(IO_TIMEOUT) ) {
			return -ERSBUSY;	//    -
		}
		udelay(IO_WAIT);
	}

	if( (ret=rs485_test_device(info)) < 0 ) {
		PDEBUG( "Test device: not found\n" );
		return ret;
	}

	//   
	copy_from_user(info->outbuffer.cache, buf, size);
	info->outbuffer.cache[size] = 0;
	info->outbuffer.size = size;

#ifndef __LPT__
	//   
	// LPT-RS485 converter   CRC

	crc_sum = crc(info->outbuffer.cache, info->outbuffer.size);
	info->outbuffer.cache[size] = crc_sum;
#endif
		//  
	ret = start_transmitter(info);
	if( (ret <= 0) || (ret > CACHE_MAXSIZE) ) return ret;
	copy_to_user((char*)buf, info->passive_inbuf.cache, ret);
	return ret;
}

static ssize_t rs485_read( struct file *fp, char *buf,
							size_t count, loff_t *offset ) {
/* 
 * Read data fifo
 */	

	char count0;
	RS485INFO	*info = fp->private_data;
	int ret;

	if( test_flags(FREAD, &info->lFlags) ) return 0;

	up_flags(FREAD, &info->lFlags);
	ret = rs485_data_fifo_read( info, buf, 1 );
	if( ret > 0 ) {
		get_user(count0,buf);	// buf in User space!!!
		if( count0 < (count-1) ) count = count0;
		else count--;
		ret = rs485_data_fifo_read( info, buf+1, count );
	}
	down_flags(FREAD, &info->lFlags);
	return ret;
}

static int rs485_ioctl(struct inode *inode, struct file *fp,
						unsigned int cmd, unsigned long arg) {

	RS485INFO *rs485 = fp->private_data;

	switch(cmd) {
		case RS_RESET:
			rs485_reset(rs485);
			return(SUCCESS);

		case RS_SET_BAUDRATE:
			rs485->baudrate = (int)arg;
			return rs485_set_baudrate(rs485);

		case RS_BLNUMBER_RESET:
			rs485_blnumber_reset(rs485);
			return(SUCCESS);

		case RS_GET_VERS:
			copy_to_user((char*)arg, RS485_UTS_RELEASE, sizeof(RS485_UTS_RELEASE));
			return(RS485_VERSION_CODE);

		case RS_GET_DATAFIFO_SIZE:
			return(SIZE_STREAM_FIFO);

		default: break;
	}
	return -ERSNOIOCTL;
}


// ---------------------------------------------------------------------

// --------    --------------------------------
struct file_operations rs485_fops = {

	open:		rs485_open,		// OPEN
	release:	rs485_close,	// CLOSE 
	read:		rs485_read,		// READ
	write:		rs485_write,	// WRITE
	ioctl:		rs485_ioctl,	// IOCTL
//	flush:		NULL			// FLUSH

};

/* ---------------------------------------------------------------
 *	   
 * --------------------------------------------------------------*/

static void rs485_free_mem1(void) {
	kfree(info_port1);
}

static void rs485_free_mem2(void) {
	kfree(info_port2);
}

static void rs485_free_irq(int port) {
//    ,   close  
	switch (port) {
		case 1 :
			rs485_stop_interrupt_card(info_port1);
			rs485_clean_irq((unsigned int) irq1);
			break;

		case 2 :
			rs485_stop_interrupt_card(info_port2);
			if(irq1 != irq2) rs485_clean_irq((unsigned int) irq2);
			break;
	}
}

static void rs485_delete_card(void) {

	release_region(io2, ADDRESS_REGION_SIZE);
	release_region(io1, ADDRESS_REGION_SIZE);
	if ( (unregister_chrdev(RS485_MAJOR, ADAPTER_ID)) !=0){
		RS_PRINT(KERN_CRIT "%s: Error unload module\n", ADAPTER_ID);
	}

	return;
}

/* -------------------------------------------------
 *      
 *  :
 * io  <base address>
 * irq <interrupt>
 * --------------------------------------------------- */
 
MODULE_AUTHOR("Andre Zaitsev, Olga Voziakova (ovoz@sai.msu.ru)");
#ifdef __REAL_TIME__
	MODULE_DESCRIPTION("RTLinux rs485 module");
#else
	MODULE_DESCRIPTION("rs485 module");
#endif
#ifdef MODULE_LICENSE
	MODULE_LICENSE("GPL");
#endif

#ifdef __CP102__
    MODULE_PARM_DESC(io1, "Base port1 I/O Moxa C102 (hex), default io1=_IO_CARD1");
    MODULE_PARM_DESC(io2, "Base port2 I/O Moxa C102 (hex), default io2=_IO_CARD2");
    MODULE_PARM_DESC(irq1, "Interrupt Moxa C102, default irq1=_IRQ_CARD");
#endif
#ifdef __COM__
    MODULE_PARM_DESC(io1, "Base COM1-port (hex), default io1=_IO_COM1");
    MODULE_PARM_DESC(io2, "Base COM2-port (hex), default io2=_IO_COM2");
    MODULE_PARM_DESC(irq1, "Interrupt COM1-port, default irq1=_IRQ_COM1");
    MODULE_PARM_DESC(irq2, "Interrupt COM2-port, default irq2=_IRQ_COM2");
#endif
#ifdef __LPT__
    MODULE_PARM_DESC(io1, "Base LPT1-port (hex), default io1=_IO_LPT1");
    MODULE_PARM_DESC(io2, "Base LPT2-port (hex), default io2=_IO_LPT2");
    MODULE_PARM_DESC(irq1, "Interrupt LPT1-port,default irq1=_IRQ_LPT1");
    MODULE_PARM_DESC(irq2, "Interrupt LPT2-port,default irq2=_IRQ_LPT2");
#endif

MODULE_PARM(io1, "i");
MODULE_PARM(io2, "i");
MODULE_PARM(irq1, "i");
MODULE_PARM(irq2, "i");

//MODULE_PARM_DESC(portnumb, "PORT NUMBER, default portnumb=_PORT_NUMB");
//MODULE_PARM(portnumb, "i");

MODULE_PARM_DESC(baudrate1, "BAUD RATE UART port1, default baudrate=_BAUDRATE_1");
MODULE_PARM_DESC(baudrate2, "BAUD RATE UART port2, default baudrate=_BAUDRATE_2");
MODULE_PARM(baudrate1, "i");
MODULE_PARM(baudrate2, "i");

EXPORT_NO_SYMBOLS;

int init_module(void) {
	int err = 0;
	int err1 = 0;
	int err2 = 0;
	char version[]=RS485_UTS_RELEASE;

//	portnumb = portnumb ? portnumb : _PORT_NUMB;	//     insmod...
	baudrate1 = baudrate1 ? baudrate1 : _BAUDRATE_1;	//	 _BAUDRATE_1,   insmod
														//	  
	baudrate2 = baudrate2 ? baudrate2 : _BAUDRATE_2;

//	PDEBUG( "PORT %d\n", portnumb );
	PDEBUG("baudrate1=%d baudrate2=%d\n", baudrate1, baudrate2);

#ifndef __PCI__
	io1 = io1 ? io1 : rs485_getIO(1);		//     insmod -
	io2 = io2 ? io2 : rs485_getIO(2);
	irq1 = irq1 ? irq1 : rs485_getIRQ(1);	//   ... 
	irq2 = irq2 ? irq2 : rs485_getIRQ(2);
#endif

	if( (err=rs485_init()) != 0 ) goto FAIL_1;

//	 

	info_port1 = (RS485INFO *) kmalloc (sizeof (RS485INFO), GFP_ATOMIC);
	if(info_port1 == NULL){
		err = -ENOMEM;
		goto FAIL_2;
	}
	
	info_port2 = (RS485INFO *) kmalloc (sizeof (RS485INFO), GFP_ATOMIC);
	if(info_port2 == NULL){
		err = -ENOMEM;
		goto FAIL_3;
	}
	
	memset((void *)info_port1, 0 , sizeof(RS485INFO)); //  
	memset((void *)info_port2, 0 , sizeof(RS485INFO)); //  

	rs485_blnumber_reset(info_port1);	//   
	rs485_blnumber_reset(info_port2);	//   

//	 	

	info_port1->base = io1;
	info_port2->base = io2;
//#ifdef __PCI__
//	if( portnumb == 2 ) info_port1->base = io + 8;	// port 2 for MOXA132
//#endif
	info_port1->usHopInp = 0;
	info_port2->usHopInp = 0;

#ifdef __PROBE__
	if( ((err1=rs485_probe(info_port1)) != 0)
		|| ((err2=rs485_probe(info_port2)) != 0) ) {
		err = err1 ? err1 : err2;
		goto FAIL_4;
	}
#endif

//    

	if( ((err1=rs485_fifo_create(info_port1)) != 0)
		|| ((err2=rs485_fifo_create(info_port2)) != 0) ) {
		err = err1 ? err1 : err2;
		goto FAIL_5;
	}

	//  

	if( (err=rs485_request_irq(irq1)) != 0 ) goto FAIL_6;
	if(irq1 != irq2) {
		if( (err=rs485_request_irq(irq2)) != 0 ) goto FAIL_7;
	}

	info_port1->baudrate = baudrate1;
	info_port2->baudrate = baudrate2;

	rs485_stop_interrupt_card(info_port1);
	rs485_stop_interrupt_card(info_port2);

#ifdef __REAL_TIME__
	RS_PRINT(KERN_INFO "%s module, REAL TIME version %s\n",
				ADAPTER_ID, version);
#else
	RS_PRINT(KERN_INFO "%s module, version %s\n",
				ADAPTER_ID, version);
#endif
	RS_PRINT(KERN_INFO "%s: install card on the base=0x%X, 0x%X and irq=%d, %d\n",
				ADAPTER_ID, io1, io2, irq1, irq2);

#ifdef __RS485_PROC__
	if( (err=rs485_create_proc()) ) goto FAIL_8;
#endif

	return(SUCCESS);

#ifdef __RS485_PROC__
FAIL_8:
#endif
	rs485_free_irq(2);
FAIL_7:
	rs485_free_irq(1);
FAIL_6:	//	request irq failed
	rs485_fifo_destroy();
FAIL_5:	//	create_fifo failed
#ifdef __PROBE__
FAIL_4:
#endif
	rs485_free_mem2();
FAIL_3: rs485_free_mem1();
FAIL_2: rs485_delete_card();
FAIL_1: return(err);
}

void cleanup_module(void){
	rs485_free_irq(2);
	rs485_free_irq(1);
	rs485_fifo_destroy();
	rs485_free_mem2();
	rs485_free_mem1();
	rs485_delete_card();
#ifdef __RS485_PROC__
	rs485_remove_proc();
#endif
	return;
}
