/* -------------------------------------
 *     
 *  
 ---------------------------------------*/
 
#define __NO_VERSION__
 
#include <linux/config.h>
 
 /*    */
#if CONFIG_MODVERSIONS==1
#include <linux/modversions.h>
#endif
 
#include <linux/module.h>
#include <linux/errno.h>

#include <linux/sched.h>
#include <asm/uaccess.h>	// put_user

#include <mcua.h>
#include <rs485.h>
#include <rserrno.h>

void rs485_fifo_flush(RS485INFO *info) {
	info->passive_inbuf.adr = 0;
	info->active_inbuf.adr = 0;
	info->fifo.data_wp = 0; info->fifo.data_rp = 0;
	atomic_set( &(info->fifo.data_filled), 0 );
}

void rs485_fifo_init(struct file* file) {
	RS485INFO *info = file->private_data;
	init_waitqueue_head( &(info->fifo.rs485_queue) );
	if( file->f_flags & O_NONBLOCK ) info->fifo.nonblock = 1;
	else info->fifo.nonblock = 0;
	rs485_fifo_flush(info);
}

int rs485_fifo_create(RS485INFO *info) {
	return(SUCCESS);
}

void rs485_fifo_destroy(void) {
	return;
}

int inline rs_min( int val1, int val2 ) {
	if( val1 <= val2 ) return val1;
	else return val2;
}

int rs485_data_fifo_test( int count, RS485INFO* info ) {
	RS485FIFO* fifo = &(info->fifo);

	if( (SIZE_STREAM_FIFO - atomic_read( &(fifo->data_filled) ))
														< count ) {
		RS_PRINT( KERN_CRIT "rs485_data_fifo_test: FIFO overflow! filled=%d count=%d\n",
					atomic_read( &(fifo->data_filled) ), count );
		return -ENOSPC;
	} else {
		return SUCCESS;
	}
}

int rs485_data_fifo_put( void* buf0, int count, RS485INFO* info ) {
/*
 * Handler's function! (race condition with fifo_read)
 */
	RS485FIFO* fifo = &(info->fifo);
	char* buf = (char*)buf0;
	int j;
	int count_writ=0;
	int count0 = count;
	int attempt = 0;

	if( (SIZE_STREAM_FIFO - atomic_read( &(fifo->data_filled) ))
													< count ) {
		RS_PRINT( KERN_CRIT "rs485_data_fifo_put: FIFO overflow 1! filled=%d count=%d\n",
					atomic_read( &(fifo->data_filled) ), count );
		return -ENOSPC;
	}

REPEAT:
//	left = (fifo->data_rp + SIZE_STREAM_FIFO -
//					fifo->data_wp ) % SIZE_STREAM_FIFO;
//	PDEBUG( "left is %i\n", left );
//	if( left == 1 ) {
//		RS_PRINT( KERN_CRIT "rs485_data_fifo_put: FIFO overflow! wp=%d rp=%d\n",
//					fifo->data_wp, fifo->data_rp );
//		return -ENOSPC;
//	}

	if( fifo->data_wp >= fifo->data_rp ) {
		count = rs_min( count, SIZE_STREAM_FIFO - fifo->data_wp );
//		if( count == left ) count--;
	} else {
// the write pointer has wrapped:
		count = rs_min( count,info->fifo.data_rp - info->fifo.data_wp );
	}
	for( j=0; j<count ; j++ ) {
		fifo->data_fifo[fifo->data_wp+j] = buf[j];
	}
	fifo->data_wp += count;
	count_writ += count;
	atomic_add( count, &(fifo->data_filled) );

	if( fifo->data_wp == SIZE_STREAM_FIFO ) {
		fifo->data_wp = 0;
	}

	if( count_writ < count0 ) {
		if( attempt > 0 ) {
			RS_PRINT( KERN_CRIT "rs485_data_fifo_put: FIFO overflow 2! attempt = %d\n",
						attempt );
			return -ENOSPC;
		}
		attempt++;
		buf = (char*)buf0+count;
		count = count0 - count;
//		PDEBUG( "put_data_fifo: repeat\n" );
		goto REPEAT;
	} else {
		return count_writ;
	}
}

ssize_t rs485_data_fifo_read( RS485INFO* info, char* buf, size_t count ) {
/*
 * NONBLOCKED read
 * copy to user space
 */
	RS485FIFO* fifo = &(info->fifo);
	int count_read = 0;
	int attempt = 0;
	int count0 = count;

	if( atomic_read( &(fifo->data_filled) ) < count ) {
		return 0;
	}

REPEAT:
//	if( fifo->data_rp  == fifo->data_wp ) {
//		return 0;
//	}

	if( fifo->data_wp > fifo->data_rp ) {
		count = rs_min( count, fifo->data_wp - fifo->data_rp );
	} else {
// the write pointer has wrapped
		count = rs_min( count, SIZE_STREAM_FIFO - fifo->data_rp );
	}
	copy_to_user( buf, fifo->data_fifo+fifo->data_rp, count);

	fifo->data_rp += count;
	count_read += count;
	atomic_sub( count, &(fifo->data_filled) );

	if( fifo->data_rp == SIZE_STREAM_FIFO ) {
		fifo->data_rp = 0;
	}

	if( count_read < count0 ) {
		if( attempt > 0 ) {
			RS_PRINT( "rs485_data_fifo_read: ERROR! attempt=%d\n",attempt );
			return 0;
		}
		attempt++;
		buf = buf+count;
		count = count0 - count;
//		PDEBUG( "read_data_fifo: repeat\n" );
		goto REPEAT;
	} else {
		return count_read;
	}
}
