#define __NO_VERSION__
#include <linux/config.h>

#if CONFIG_MODVERSIONS==1
#include <linux/modversions.h>
#endif

#include <linux/module.h>

#include <asm/irq.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/time.h>

#include <mcua.h>
#include "rs485.h"

#ifdef __CP132__
	#include <moxa132.h>
#endif
#ifdef __CP102__
	#include <moxa102.h>
#endif
#ifdef __COM__
	#include <com.h>
#endif
#ifdef __LPT__
	#include <lpt.h>
#endif

extern  inline void reciv_data(RS485INFO *info);
extern RS485INFO *info_port1, *info_port2;
extern int irq1, irq2;

static struct timeval start_waiting;

int iround(float a) {
	int i;
	if(a >= INT_MAX) return(INT_MAX);
	if(a <= INT_MIN) return(INT_MIN);
	i = (int) a;
	if(a < (float)i + 0.5) return i;
	return(i+1); 
}

inline void rs485_gettime(rs485_time_t *tm) {
	do_gettimeofday(tm);
	return;
}

inline void rs485_printtime(rs485_time_t *tm) {
	RS_PRINT( "rs485-time> %ld s %ld usec\n",
						tm->tv_sec, tm->tv_usec );
	return;
}

inline void rs485_printtm_interval(rs485_time_t *tm_start,
										rs485_time_t *tm_end) {
	RS_PRINT( "rs485-dtime> %ld usec\n",
					(tm_end->tv_sec - tm_start->tv_sec)*1000000L+
					(tm_end->tv_usec - tm_start->tv_usec) );
	return;
}

inline void rs485_set_starttime(void) {
	do_gettimeofday( &start_waiting );
}

inline int rs485_timeout(long interval) {	// usec
	struct timeval tv;

	do_gettimeofday(&tv);
/*
	long usec = (tv.tv_sec-start->tv_sec)*1000000L
					 + (tv.tv_usec-start->tv_usec);
	PDEBUGG( "%ld : %ld - %ld : %ld => dusec=%ld\n",
				tv.tv_sec,
				tv.tv_usec,
				start->tv_sec,
				start->tv_usec,
				usec );
*/
	if( ( (tv.tv_sec - start_waiting.tv_sec)*1000000L
					 + (tv.tv_usec - start_waiting.tv_usec) )
											> interval ) {
		PDEBUGG( "Timeout!!!\n" );
		return 1;
	} else return 0;
}

static void rs485_interrupt(int irq, void *dev_id, struct pt_regs *regs) {
//    
	RS485INFO *info_port;

//	PDEBUG("rs485_interrupt! irq=%d\n", irq);
	if(irq1 == irq2) {
//		raspoznavanie porta
		info_port = info_port1;	// ili 2
	} else {
		if(irq == irq1) info_port = info_port1;
		else info_port = info_port2;
	}
	reciv_data(info_port);
	return;
}

int rs485_request_irq(unsigned int irq) {
	unsigned long flags;
	int err=0;

	save_flags(flags); cli();
	if( (err=request_irq(irq, rs485_interrupt,
				SA_INTERRUPT, ADAPTER_ID,NULL)) != 0 ) {
//				0, ADAPTER_ID,NULL)) != 0 ) {
		RS_PRINT(KERN_CRIT "rs485: request irq %d failed\n",irq);
	}
	sti(); restore_flags(flags);
	return err;
}

void rs485_clean_irq(unsigned int irq) {
	unsigned long flags;
	save_flags(flags); cli();
	free_irq(irq,NULL);
	sti(); restore_flags(flags);
}

static unsigned long flags;
static int is_flags_saved = 0;

inline void rs485_stop_interrupts(void) {

// disable all interrupts

	save_flags(flags);
	is_flags_saved = 1;
	cli();
}

inline void rs485_allow_interrupts(void) {

// enable interrupts

	sti();
	if(is_flags_saved) restore_flags(flags);
}
