#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <pthread.h>
#include <signal.h>

#include "../include/mcua.h"

#define FLAGS_WAIT_STATUS	1
#define FLAGS_READ_BLOCK	2
#define FLAGS_DOUT_DATA 	4

pthread_mutex_t lockmsg  = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t lockdata = PTHREAD_MUTEX_INITIALIZER;

//static unsigned char msgbuf[40];
static unsigned char *msgdata;

pthread_attr_t	atrib;
pthread_t		th_stat, th_timer, th_data;
static int id_mcua;
static int id_proc;
//static int id_out, id_stat=0;
//static int id_data;
FILE	*fileinp;

const int counter = 0x1;
const int N_BASE_TIME = 1;
#define	DATABUF_MAXSIZE	4096

void *read_data(void *arg) {	//   
	int yy, t;
	int count = 0;
	
	printf("start read data...\n" );
	while(1) {
		yy = 0;
		while( yy < 32 ) {
			count = read( id_mcua, msgdata, DATABUF_MAXSIZE );
			if( count == 0 ) {
				usleep( 10000 );
				continue;
			}
			for( t=0; t<=count; t++ ) {
				printf("%02X ", msgdata[t]);
			}
			printf("\n");
			yy++;
		}
		pthread_mutex_unlock(&lockdata);
	}
pthread_exit(0);
}


void terminate_program(void){
	close(id_mcua);
	close(id_proc);
	
	pthread_kill(th_data, SIGINT);
	pthread_join(th_data, NULL);
	pthread_mutex_destroy(&lockdata);
	free(msgdata);	
}

static unsigned char buf[CACHE_MAXSIZE];

int main (int argc, char **argv){
	struct  sched_param param;
	int retval;
	int i;
	msgdata = (char *) malloc(DATABUF_MAXSIZE);

	
	if (pthread_attr_init(&atrib) == -1){	//  
		perror("Create attributes");
		exit(1);
	}	
	
	param.sched_priority = 20;				//   
	pthread_attr_setschedparam(&atrib, &param);
	pthread_attr_setschedpolicy(&atrib, PTHREAD_SCOPE_PROCESS);
	if( (id_mcua=open( "/dev/mcua0", O_RDWR | O_NONBLOCK )) < 0 ) { // 
		perror("Open mcua");
		exit(3);
	}

	id_proc = open( "/proc/moxa/mcua0/device", O_RDWR );
	write( id_proc,"AAA",3 );

	ioctl(id_mcua, RS_BLNUMBER_RESET);	
	
	pthread_mutex_lock(&lockdata);
	if( pthread_create(&th_data, &atrib, read_data, NULL) == -1 ) {
		perror("Create thread's");
		exit(2);
	}
	
	atexit( terminate_program);
	
	/*-------------------------------------- *
	 *    			 *
	 *-------------------------------------- */	
/*
		buf[0] = counter;
		buf[1] = 0x52;	// Set expos. 4 msec
		buf[1] = 0x65;	//
		buf[1] = 0xE;	//
		
//		ioctl(id_mcua, READ_STAT_FIFO);	
		retval = write( id_mcua, buf, 4 ); // 
		if( retval < 0 ) {
			perror("txd: ");
			exit(1);
		} else if( retval > CACHE_MAXSIZE ) {
			printf("Signal : 0x%X\n", retval);
		} else {
//			ioctl(id_mcua, READ_STAT_FIFO);	
			msgbuf[0]=STAT;
			yy = read(id_mcua, msgbuf, retval);
			for( i=0; i<yy; i++){
				printf("%X ", msgbuf[i]);
			}
			printf("\n");
		}	 
*/	 
		buf[0] = counter;
		buf[1] = 0x8B;	// Induce off
		
		retval = write(id_mcua, buf, 2); // 
		if( retval < 0 ) {
			perror("txd: ");
			exit(1);
		} else if( retval > CACHE_MAXSIZE ) {
			printf("Signal : 0x%X\n", retval);
		} else {
			for( i=1; i<retval; i++) {
				printf("%X ", buf[i]);
			}
			printf("\n");
		}	 
	 
		buf[0] = counter;
		buf[1] = 0x83;	// Master on
		
		retval = write(id_mcua,buf,2); // 
		if( retval < 0 ) {
			perror("txd: ");
			exit(1);
		} else if( retval > CACHE_MAXSIZE ) {
			printf("Signal : 0x%X\n", retval);
		} else {
			for( i=1; i<retval; i++) {
				printf("%X ", buf[i]);
			}
			printf("\n");
		}	 
		buf[0] = counter;
		buf[1] = 0x88;	// Active on
		
		retval = write( id_mcua, buf, 2); // 
		if( retval < 0 ) {
			perror("txd: ");
			exit(1);
		} else if( retval > CACHE_MAXSIZE ) {
			printf("Signal : 0x%X\n", retval);
		} else {
			for( i=1; i<retval; i++) {
				printf("%X ", buf[i]);
			}
			printf("\n");
		}	 
	 
	 for( i=0; i<N_BASE_TIME; i++ ) {

		buf[0] = counter;
		buf[1] = 0x86;		
		retval = write( id_mcua, buf, 2 ); // 
						
		if( retval < 0 ) {
			perror("txd: ");
			exit(1);
		} else if( retval > CACHE_MAXSIZE ) {
			printf("Signal : 0x%X\n", retval);
		} else {
			for( i=1; i<retval; i++) {
				printf("%02X ", buf[i]);
			}
			printf("\n");
		}	 				
		pthread_mutex_lock(&lockdata);
	 
	 }
	 
exit(0);
}
